// LED Pattern Demo - Part 4, Orthogonal
//
// An example with brightness control orthogonal to pattern control
//
// This sketch uses:
//  - Timer 3 to manage brightness through a bit-bang PWM
//  - The LED pattern manager to change the patterns
//  - A driver function changes the input / output mode of the pin
//    to control whether or not it should be "ON"


// How does this sketch work?
// ==========================
//
// It uses both the port pin direction (i.e. is the pin configured to be an
// OUTPUT or an INPUT), and the actual state (i.e. is it ON or OFF) to achieve
// orthogonal control over the ON-ness, and the brightness (i.e. the percent of
// time spent ON when is should be on).
//
// I'm not sure if this is a good idea, but it serves the purpose for this
// sketch.
//
// The Timer 3 ISR handlers setup a PWM where the duty of the PWM is controlled
// by the "brightness" value (as a fraction over 255...the PWM rollover).
//
// When the pin should be "ON", the LED driver function switches the pin to be
// an OUTPUT -> thus outputting the PWM from timer 3.
//
// When the pin should be "OFF", the LED driver function switches the pin to be
// an INPUT -> restricting the PWN, turning the pin into a high-impedance pin
// (I think...I didn't check). Really we should make sure that the pin is
// pulled low/high, whatever we need to prevent current flowing...but for this
// sketch is seems to work.


/**********************************************
  START of LED Management Module
 **********************************************/

enum {
  // The maximum number of patterns that can be registered (to avoid memory
  // management)
  MAX_PATTERNS = 5,
  // Set prescale to 64
  PRESCALE_VALUE = (1 << CS10) | (1 << CS11),
  // Preload timer value to give 8Hz rollover:
  //  - 64 prescaler -> 16000000 Hz / 64 / 8Hz = 31250
  TIMER_PRELOAD_VALUE = 31250,
};

/**
  The function signature called by the Led Manager to drive the "led" to a
  state.

  @param color The color that the LED should be set to
  @param arg   The user provided value when the pattern was registered
*/
typedef int (*LedManager_UpdateFunc)(char color, void *arg);

// An internal type to store the pattern information
typedef struct {
  // The external "value" that correspondes to this pattern. Storing this allows
  // the "outside" caller to determine what the values provided to
  // LedMgr_activate() are
  long value;
  // A pointer to a string (hopefully immutable / in flash) that represents this
  // pattern
  char *pattern;
  // The driver function to call to update the LED module
  LedManager_UpdateFunc handler;
  // The argument to be handed back to the driver function
  void *arg;
} pattern_bucket;

// Our internal store of the available patterns
static pattern_bucket patterns[MAX_PATTERNS] = {0};
// The active pattern (NB: Not necessarily set)
static pattern_bucket *current_pattern = NULL;

/*
   This ISR is responsible for driving the LED change. Possibly a little risky
   in an ISR, but should be ok for this sketch.
*/
ISR(TIMER1_OVF_vect) // interrupt service routine for overflow
{
  // Reset value to make sure this happens again in 125ms
  TCNT1 = TIMER_PRELOAD_VALUE;

  static int counter = 0;
  if (current_pattern != NULL) {
    int len = strlen(current_pattern->pattern);
    char val = current_pattern->pattern[counter % len];
    current_pattern->handler(val, current_pattern->arg);
  }
  counter++;
}

/*
   Activate the pattern registered with the provided value.

   If the value does not match any registered pattern, disable the active
   pattern
*/
void LedMgr_activate(long value) {
  for (int i = 0; i < MAX_PATTERNS; i++) {
    if (patterns[i].value == value) {
      current_pattern = &patterns[i];
      return;
    }
  }
  current_pattern = NULL; // no patterns match, disable current pattern
}

/*
   Adds a new pattern

   @param value   An external value to use as the "key" for activating this
   pattern
   @param pattern The pattern, as a string, in an encoding understood by the
   provided UpdateFunc
   @param handler A function that can respond to each character in the pattern
   and drive the "LED" correctly
   @param arg     An external parameter to pass through to the handler

   @retval 0  if successful
   @retval -1 if unable to add the pattern
*/
int LedMgr_addPattern(long value, char const *const pattern,
                      LedManager_UpdateFunc handler, void *arg) {
  for (int i = 0; i < MAX_PATTERNS; i++) {
    pattern_bucket *p = &patterns[i];
    // If the current bucket is empty, use it
    if (p->pattern == NULL) {
      p->value = value;
      p->pattern = pattern;
      p->handler = handler;
      p->arg = arg;
      return 0;
    }
    // If this bucket is already used AND the value is the same, overwrite it
    if (p->value == value) {
      p->pattern = pattern;
      p->handler = handler;
      p->arg = arg;
      return 0;
    }
  }
  // Not enough buckets left
  return -1;
}

/**********************************************
  END of LED Management Module
 **********************************************/

/**********************************************
  START of Application
 **********************************************/

enum State {
  FOUR_TO_THE_FLOOR,
  SWUNG_BEAT,
  THREE_FOUR,
};

// Constants for our pins
enum {
  RED_PIN = 11,
  GREEN_PIN = 3,
  YELLOW_PIN = 5,

  LOOP_DELAY = 3000,
  STEP = 5,
};

// A type to capture a three-led module, comprised of red, green and yellow LEDs
typedef struct {
  int red;
  int green;
  int yellow;
} tricolour_led_module;

// The ascii characters are arranged in a way such that the uppercase and
// lowercase characters only have the 6th bit set (to mean "lowercase").
// http://www.asciitable.com makes this quite clear. We are going to utilise
// this to simplify our character checks by first forcing the provided char
// to be lowercase (by ORing with this bitmask, i.e. by setting the 6th bit).
enum {
  LOWERCASE_BIT = 0x20,
};

/**
  Drive the provided pin to the provided pattern element, but ignore the case
  of each char.

  This controls the ON-ness by modulating the pin direction.

  @param color  the pattern element
  @param arg    a pointer to a tricolour_led_module
*/
int tricolorDrive(char color, void *arg) {
  tricolour_led_module *pin = (tricolour_led_module *)arg;
  pinMode(pin->red,    (color | LOWERCASE_BIT) == 'r' ? OUTPUT : INPUT);
  pinMode(pin->green,  (color | LOWERCASE_BIT) == 'g' ? OUTPUT : INPUT);
  pinMode(pin->yellow, (color | LOWERCASE_BIT) == 'y' ? OUTPUT : INPUT);
}

void setup() {
  Serial.begin(9600);

  pinMode(RED_PIN, OUTPUT);
  pinMode(GREEN_PIN, OUTPUT);
  pinMode(YELLOW_PIN, OUTPUT);

  static tricolour_led_module leds = {
    .red = RED_PIN, .green = GREEN_PIN, .yellow = YELLOW_PIN
  };

  // Register the patterns against the states
  LedMgr_addPattern(FOUR_TO_THE_FLOOR, "R--g--g--g--", tricolorDrive, &leds);
  LedMgr_addPattern(SWUNG_BEAT,        "RRR   r   G ", tricolorDrive, &leds);
  LedMgr_addPattern(THREE_FOUR,        "RrrrGgggYyyy", tricolorDrive, &leds);

  noInterrupts(); // disable all interrupts
  // Setup Timer1 to create an 8Hz ISR trigger (if CPU Freq is 16 MHz) for LED Manager
  TCCR1A = 0;
  TCCR1B = 0;
  TCNT1 = TIMER_PRELOAD_VALUE; // preload timer
  TCCR1B |= PRESCALE_VALUE;
  TIMSK1 |= (1 << TOIE1); // enable timer overflow interrupt ISR

  // Waveform Generation
  // ===================
  //  Mode  Mode of Operation  TOP       Update of OCRnx at    TOVn Flag Set on
  //    3 - Fast PWM, 8-bit    0x00FF          TOP                  BOTTOM

  // 14.10.2 Timer/Counter3 Control Register A – TCCR3A
  // 7      6      5      4      3      2      1      0
  // COM3A1 COM3A0 COM3B1 COM3B0 COM3C1 COM3C0 WGM31  WGM30
  // 0      0      0      0      0      0      0      1
  TCCR3A = 0x01;
  // 14.10.4 Timer/Counter3 Control Register B – TCCR3B
  // 7      6      5      4      3      2      1      0
  // ICNC3  ICES3  –      WGM33  WGM32  CS32   CS31   CS30
  // 0      0      0      0      1      1      0      1
  TCCR3B = 0x0D;

  // Enable Overflow and Output Compare 1 interrupt
  TIMSK3 |= (1 << TOIE3) | (1 << OCIE3A);
  OCR3AH = 0;
  OCR3AL = 0;

  interrupts();           // enable all interrupts
}

static int brightness = 20;
static int dir = 5;

ISR(TIMER3_COMPA_vect) {
  digitalWrite(RED_PIN, LOW);
  digitalWrite(GREEN_PIN, LOW);
  digitalWrite(YELLOW_PIN, LOW);
}

ISR(TIMER3_OVF_vect) {
  digitalWrite(RED_PIN, brightness > 0);
  digitalWrite(GREEN_PIN, brightness > 0);
  digitalWrite(YELLOW_PIN, brightness > 0);
  OCR3AL = brightness;
}

void loop() {
  delay(LOOP_DELAY);
  static int loops = 0;

  // Every loop switch the brightness
  if (brightness == 250) {
    brightness = 10;
  } else {
    brightness = 250;
  }

  Serial.print("Brightness:");
  Serial.print(brightness);
  Serial.println("");

  // Every 3 loops, switch the pattern
  if (loops % 3 == 0) {
    static long nextBeat = 0;
    switch (nextBeat) {
      case 0:
        LedMgr_activate(FOUR_TO_THE_FLOOR);
        Serial.println("four to the floor");
        nextBeat = 1;
        break;
      case 1:
        LedMgr_activate(SWUNG_BEAT);
        Serial.println("feel the swing");
        nextBeat = 2;
        break;
      default:
        LedMgr_activate(THREE_FOUR);
        Serial.println("3/4 beat");
        nextBeat = 0;
        break;
    }
  }

  loops++;
}
